// SPDX-License-Identifier: BSD-2-Clause

// This code is part of the sfizz library and is licensed under a BSD 2-clause
// license. You should have receive a LICENSE.md file along with the code.
// If not, contact the sfizz maintainers at https://github.com/sfztools/sfizz

#pragma once
#include <functional>
#include <cstddef>

namespace sfz { class ModKey; }

namespace std {
    template <> struct hash<sfz::ModKey> {
        size_t operator()(const sfz::ModKey &key) const;
    };
}
